/*
 * Program to receive packets for Multicast groups.
 * 
 *
 * Written by Polly Huang, University of Southern California, July 1995.
 * Bug fixes and small modifications by Pavlin Radoslavov, May 1997
 */

#define MULTICAST


#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <netinet/in.h>

#define INFINITE        24*60*60*1000000

typedef struct
{
  u_short seq_num;
  unsigned char mdestadd[4];
  u_short port_no;
  u_short length;
  char data[10000];
} Packet;

main( argc, argv )
     int argc;
     char **argv;
{
  int so;
  char line[80];
  char *lineptr;
  struct ip_mreq imr;   /* mutilcast join/leave address */
  struct ifreq ifr;     /* not known yet */
  unsigned i1, i2, i3, i4, g1, g2, g3, g4;
  unsigned e1, e2, e3, e4, e5, e6; /* ethernet add */
  struct in_addr dest_group; /* in_addr + port# = sockaddr_in */
  struct sockaddr_in dest;
  u_short port_no;
  int addlen;           /* address length for bind() */
  Packet packet;
  int n;                /* return value of system call */
  long received_bytes;

  bzero(&imr, sizeof(imr));
  bzero(&ifr, sizeof(ifr));
  bzero(&dest_group, sizeof(dest_group));
  bzero(&dest, sizeof(dest));
  bzero(&packet, sizeof(packet));


  if( (so = socket( AF_INET, SOCK_DGRAM, 0 )) == -1)
    {
      perror( "can't open socket" );
      exit( 1 );
    }
  dest.sin_family = AF_INET;

  printf( "multicast receiving test program; " );
  printf( "enter ? for list of commands\n" );
  
  while( fgets( line, 79, stdin ) != NULL )
    {
      lineptr = line;
      while( *lineptr == ' ' || *lineptr == '\t' ) ++lineptr;

      switch( *lineptr )
	{
	case '?':
	  {
	    printf( "%s%s%s%s",
		   " j g.g.g.g i.i.i.i              - join  IP  multicast group \n",
		   " l g.g.g.g i.i.i.i              - leave IP  multicast group \n",
		   " r g.g.g.g port#                - receive from multicast interface \n",
		   " q                              - quit                      \n\n" );
	    break;
	  }
	  
	case 'r':
	  {
	    ++lineptr;
	    while( *lineptr == ' ' || *lineptr == '\t' ) ++lineptr;
	    if( (n = sscanf( lineptr, "%u.%u.%u.%u %hd",
			    &g1, &g2, &g3, &g4, &port_no)) != 5 )
	      {
		printf( "bad args\n" );
		break;
	      }
	    dest_group.s_addr = (g1<<24) | (g2<<16) | (g3<<8) | g4;
	    dest_group.s_addr = htonl(dest_group.s_addr);
/* TODO: Doesn't work for Solaris 5.4	    
	    dest.sin_addr = dest_group;
*/
	    dest.sin_port = htons(port_no);
	    addlen = sizeof (struct sockaddr_in);
	    if (bind (so, (struct sockaddr *)&dest, addlen) < 0)
	      printf ("Error in binding socket\n");
	    else
	      printf ("Bind socket\n");

	    do {
	      if((received_bytes =
		  recvfrom(so, (char *)&packet, sizeof(packet), 0, 
			NULL, &addlen)) < 0) {
		perror("Error in receiving pkt");
	      } else {
		  if (received_bytes == ntohs(packet.length))
		      printf ("Receive pkt # %d, for mgroup %u.%u.%u.%u port %hd, length=%d\n", ntohs(packet.seq_num), packet.mdestadd[0], packet.mdestadd[1], packet.mdestadd[2], packet.mdestadd[3], ntohs(packet.port_no), ntohs(packet.length));
		  else
		      printf ("Receive pkt # %d, for mgroup %u.%u.%u.%u port %hd, length=%d, but MUST BE %d\n", ntohs(packet.seq_num), packet.mdestadd[0], packet.mdestadd[1], packet.mdestadd[2], packet.mdestadd[3], ntohs(packet.port_no), received_bytes, ntohs(packet.length));
	      }
	    } while (1);
/*	    }while (ntohs(packet.seq_num) != 9999); */
	    break;
	  } 

	case 'j':
	  {
	    ++lineptr;
	    while( *lineptr == ' ' || *lineptr == '\t' ) ++lineptr;
	    if( (n = sscanf( lineptr, "%u.%u.%u.%u %u.%u.%u.%u %u",
			    &g1, &g2, &g3, &g4, &i1, &i2, &i3, &i4 )) != 8 )
	      {
		printf( "bad args\n" );
		break;
	      }
	    imr.imr_multiaddr.s_addr = (g1<<24) | (g2<<16) | (g3<<8) | g4;
	    imr.imr_multiaddr.s_addr = htonl(imr.imr_multiaddr.s_addr);
	    imr.imr_interface.s_addr = (i1<<24) | (i2<<16) | (i3<<8) | i4;
	    imr.imr_interface.s_addr = htonl(imr.imr_interface.s_addr);
	    if( setsockopt( so, IPPROTO_IP, IP_ADD_MEMBERSHIP,
			   (char *)&imr, sizeof(struct ip_mreq) ) == -1 )
	      perror( "can't join group" );
	    else printf( "group joined\n" );
	    break;
	  }

	case 'l':
	  {
	    ++lineptr;
	    while( *lineptr == ' ' || *lineptr == '\t' ) ++lineptr;
	    if( (n = sscanf( lineptr, "%u.%u.%u.%u %u.%u.%u.%u %u",
			    &g1, &g2, &g3, &g4, &i1, &i2, &i3, &i4 )) != 8 )
	      {
		printf( "bad args\n" );
		break;
	      }
	    imr.imr_multiaddr.s_addr = (g1<<24) | (g2<<16) | (g3<<8) | g4;
	    imr.imr_multiaddr.s_addr = htonl(imr.imr_multiaddr.s_addr);
	    imr.imr_interface.s_addr = (i1<<24) | (i2<<16) | (i3<<8) | i4;
	    imr.imr_interface.s_addr = htonl(imr.imr_interface.s_addr);
	    if( setsockopt( so, IPPROTO_IP, IP_DROP_MEMBERSHIP,
			   (char *)&imr, sizeof(struct ip_mreq) ) == -1 )
	      perror( "can't leave group" );
	    else printf( "group left\n" );
	    break;
	  }     

	case 'q': exit( 0 );
	  
	case '\n': break;
	  
	default:
	  {
	    printf( "bad command\n" );
	    break;
	  }
	}
    }
}

